from typing import Optional, overload, Union, TYPE_CHECKING

from klotho.aws.network import Network
from klotho.construct import (
    ConstructOptions,
    get_construct_args_opts,
    Construct,
    Binding,
    add_binding,
)
from klotho.output import Input, Output, MappingInput
from klotho.runtime_util import get_default_construct
from klotho.type_util import set_field, get_field, get_output

if TYPE_CHECKING:
    from klotho.aws.bucket import Bucket
    from klotho.aws.postgres import Postgres

BindingType = Union[
    Binding["Bucket"], "Bucket",
    Binding["Postgres"], "Postgres",
]


class ContainerArgs:
    def __init__(
        self,
        bindings: Optional[list[BindingType]] = None,
        context: Optional[Input[str]] = None,
        cpu: Optional[Input[int]] = None,
        dockerfile: Optional[Input[str]] = None,
        enable_execute_command: Optional[Input[bool]] = None,
        environment_variables: Optional[MappingInput[str]] = None,
        image: Optional[Input[str]] = None,
        memory: Optional[Input[int]] = None,
        network: Optional[Network] = None,
        port: Optional[Input[int]] = None,
    ):
        if bindings is not None:
            set_field(self, "bindings", bindings)
        if context is not None:
            set_field(self, "context", context)
        if cpu is not None:
            set_field(self, "cpu", cpu)
        if dockerfile is not None:
            set_field(self, "dockerfile", dockerfile)
        if enable_execute_command is not None:
            set_field(self, "enable_execute_command", enable_execute_command)
        if environment_variables is not None:
            set_field(self, "environment_variables", environment_variables)
        if image is not None:
            set_field(self, "image", image)
        if memory is not None:
            set_field(self, "memory", memory)
        if network is not None:
            set_field(self, "network", network)
        if port is not None:
            set_field(self, "port", port)

    @property
    def image(self) -> Input[str] | None:
        return get_field(self, "image")

    @image.setter
    def image(self, value: Input[str]) -> None:
        set_field(self, "image", value)

    @property
    def cpu(self) -> Optional[Input[int]]:
        return get_field(self, "cpu")

    @cpu.setter
    def cpu(self, value: Optional[Input[int]]) -> None:
        set_field(self, "cpu", value)

    @property
    def memory(self) -> Optional[Input[int]]:
        return get_field(self, "memory")

    @memory.setter
    def memory(self, value: Optional[Input[int]]) -> None:
        set_field(self, "memory", value)

    @property
    def context(self) -> Optional[Input[str]]:
        return get_field(self, "context")

    @context.setter
    def context(self, value: Optional[Input[str]]) -> None:
        set_field(self, "context", value)

    @property
    def dockerfile(self) -> Optional[Input[str]]:
        return get_field(self, "dockerfile")

    @dockerfile.setter
    def dockerfile(self, value: Optional[Input[str]]) -> None:
        set_field(self, "dockerfile", value)

    @property
    def network(self) -> Network | None:
        return get_field(self, "network")

    @network.setter
    def network(self, value: Optional[Network]) -> None:
        set_field(self, "network", value)

    @property
    def enable_execute_command(self) -> Optional[Input[bool]]:
        return get_field(self, "enable_execute_command")

    @enable_execute_command.setter
    def enable_execute_command(self, value: Optional[Input[bool]]) -> None:
        set_field(self, "enable_execute_command", value)

    @property
    def environment_variables(self) -> Optional[MappingInput[str]]:
        return get_field(self, "environment_variables")

    @environment_variables.setter
    def environment_variables(self, value: Optional[MappingInput[str]]) -> None:
        set_field(self, "environment_variables", value)

    @property
    def port(self) -> Optional[Input[int]]:
        return get_field(self, "port")

    @port.setter
    def port(self, value: Optional[Input[int]]) -> None:
        set_field(self, "port", value)


class Container(Construct):

    @overload
    def __init__(
        self, name: str, args: ContainerArgs, opts: Optional[ConstructOptions] = None
    ): ...

    @overload
    def __init__(
        self,
        name: str,
        bindings: Optional[list[BindingType]] = None,
        context: Optional[Input[str]] = None,
        cpu: Optional[Input[int]] = None,
        dockerfile: Optional[Input[str]] = None,
        enable_execute_command: Optional[Input[bool]] = None,
        environment_variables: Optional[MappingInput[str]] = None,
        image: Optional[Input[str]] = None,
        memory: Optional[Input[int]] = None,
        network: Optional[Network] = None,
        port: Optional[Input[int]] = None,
        opts: Optional[ConstructOptions] = None,
    ): ...

    def __init__(self, name: str, *args, **kwargs):
        construct_args, opts = get_construct_args_opts(ContainerArgs, *args, **kwargs)
        if construct_args is not None:
            self._internal_init(name, opts=opts, **construct_args.__dict__)
        else:
            self._internal_init(name, *args, **kwargs)

    def _internal_init(
        self,
        name: str,
        context: Optional[Input[str]] = None,
        cpu: Optional[Input[int]] = None,
        dockerfile: Optional[Input[str]] = None,
        enable_execute_command: Optional[Input[bool]] = None,
        environment_variables: Optional[MappingInput[str]] = None,
        image: Optional[Input[str]] = None,
        memory: Optional[Input[int]] = None,
        network: Optional[Network] = None,
        port: Optional[Input[int]] = None,
        bindings: Optional[list[BindingType]] = None,
        opts: Optional[ConstructOptions] = None,
    ):
        if network is None:
            network = get_default_construct("aws", Network)

        super().__init__(
            name,
            construct_type="klotho.aws.Container",
            properties={
                "Context": context,
                "Cpu": cpu,
                "Dockerfile": dockerfile,
                "EnableExecuteCommand": enable_execute_command,
                "EnvironmentVariables": (
                    Output.from_mapping(environment_variables)
                    if environment_variables
                    else None
                ),
                "Image": image,
                "Memory": memory,
                "Network": network,
                "Port": port,
            },
            bindings=bindings,
            opts=opts,
        )

    @property
    def load_balancer_url(self) -> Output[str]:
        return get_output(self, "LoadBalancerUrl", str)

    def bind(self, binding: BindingType) -> None:
        add_binding(self, binding)



from typing import Optional, overload, List, Dict

from klotho.construct import ConstructOptions, get_construct_args_opts, Construct
from klotho.output import Input, MappingInput, Output
from klotho.type_util import set_field, get_field, get_output


class DynamoDBArgs:
    """Arguments for configuring a DynamoDB table."""

    def __init__(
                 self,
                 attributes: Input[List[MappingInput[str]]],
                 hash_key: Input[str],
                 billing_mode: Optional[Input[str]] = None,
                 range_key: Optional[Input[str]] = None,
                 global_secondary_indexes: Optional[Input[List[MappingInput[str]]]] = None, 
                 local_secondary_indexes: Optional[Input[List[MappingInput[str]]]] = None,
                 tags: Optional[Input[MappingInput[str]]] = None
                ):
        set_field(self, "attributes", attributes)
        set_field(self, "hash_key", hash_key)
        if billing_mode is not None:
            set_field(self, "billing_mode", billing_mode)
        if range_key is not None:
            set_field(self, "range_key", range_key)
        if global_secondary_indexes is not None:
            set_field(self, "global_secondary_indexes", global_secondary_indexes)
        if local_secondary_indexes is not None:
            set_field(self, "local_secondary_indexes", local_secondary_indexes)
        if tags is not None:
            set_field(self, "tags", tags)

    def _get_property(self, name: str):
        return get_field(self, name)

    def _set_property(self, name: str, value):
        set_field(self, name, value)

    @property
    def attributes(self) -> Optional[Input[List[MappingInput[str]]]]:
        return self._get_property("attributes")

    @attributes.setter
    def attributes(self, value: Input[List[MappingInput[str]]]) -> None:
        self._set_property("attributes", value)

    @property
    def billing_mode(self) -> Optional[Input[str]]:
        return self._get_property("billing_mode")

    @billing_mode.setter
    def billing_mode(self, value: Optional[Input[str]]) -> None:
        self._set_property("billing_mode", value)

    @property
    def hash_key(self) ->Optional[Input[str]]:
        return self._get_property("hash_key")

    @hash_key.setter
    def hash_key(self, value: Input[str]) -> None:
        self._set_property("hash_key", value)

    @property
    def range_key(self) -> Optional[Input[str]]:
        return self._get_property("range_key")

    @range_key.setter
    def range_key(self, value: Optional[Input[str]]) -> None:
        self._set_property("range_key", value)

    @property
    def global_secondary_indexes(self) -> Optional[Input[List[MappingInput[str]]]]:
        return self._get_property("global_secondary_indexes")

    @global_secondary_indexes.setter
    def global_secondary_indexes(self, value: Optional[Input[List[MappingInput[str]]]]) -> None:
        self._set_property("global_secondary_indexes", value)

    @property
    def local_secondary_indexes(self) -> Optional[Input[List[MappingInput[str]]]]:
        return self._get_property("local_secondary_indexes")

    @local_secondary_indexes.setter
    def local_secondary_indexes(self, value: Optional[Input[List[MappingInput[str]]]]) -> None:
        self._set_property("local_secondary_indexes", value)

    @property
    def tags(self) -> Optional[Input[MappingInput[str]]]:
        return self._get_property("tags")

    @tags.setter
    def tags(self, value: Optional[Input[MappingInput[str]]]) -> None:
        self._set_property("tags", value)

class DynamoDB(Construct):
    """Represents a DynamoDB table construct in AWS."""

    @overload
    def __init__(
        self, name: str, args: DynamoDBArgs, opts: Optional[ConstructOptions] = None
    ): ...

    @overload
    def __init__(
        self,
        name: str,
        attributes: Input[List[MappingInput[str]]],
        hash_key: Input[str],
        billing_mode: Optional[Input[str]] = None,
        range_key: Optional[Input[str]] = None,
        global_secondary_indexes: Optional[Input[List[MappingInput[str]]]] = None,  
        local_secondary_indexes: Optional[Input[List[MappingInput[str]]]] = None,  
        tags: Optional[MappingInput[str]] = None,
        opts: Optional[ConstructOptions] = None,
    ): ...

    def __init__(self, name: str, *args, **kwargs):
        construct_args, opts = get_construct_args_opts(DynamoDBArgs, *args, **kwargs)
        if construct_args is not None:
            self._internal_init(name,  opts=opts, **construct_args.__dict__)
        else:
            self._internal_init(name, *args, **kwargs)

    def _internal_init(
        self,
        name: str,
        attributes: Input[List[MappingInput[str]]],
        hash_key: Input[str],
        billing_mode: Optional[Input[str]] = None,
        range_key: Optional[Input[str]] = None,
        global_secondary_indexes: Optional[Input[List[MappingInput[str]]]] = None,  
        local_secondary_indexes: Optional[Input[List[MappingInput[str]]]] = None,  
        tags: Optional[Input[MappingInput[str]]] = None,
        opts: Optional[ConstructOptions] = None,
    ):
        """Internal initializer for DynamoDB."""
        if billing_mode is None:
            billing_mode = "PAY_PER_REQUEST"

        super().__init__(
            name,
            construct_type="klotho.aws.DynamoDB",
            properties={
                "Attributes": attributes,
                "BillingMode": billing_mode,
                "HashKey": hash_key,
                "RangeKey": range_key,
                "GlobalSecondaryIndexes": global_secondary_indexes,  
                "LocalSecondaryIndexes": local_secondary_indexes,   
                "Tags": tags,
            },
            opts=opts,
        )

    @property
    def table_name(self) -> Output[str]:
        """The name of the DynamoDB table."""
        return get_output(self, "TableName", str)

    @property
    def table_arn(self) -> Output[str]:
        """The Amazon Resource Name (ARN) of the DynamoDB table."""
        return get_output(self, "TableArn", str)



from typing import Optional, overload, Any

from klotho.construct import (
    ConstructOptions,
    get_construct_args_opts,
    Construct,
    Binding,
)
from klotho.output import Input, Output
from klotho.type_util import set_field, get_field, get_output


class BucketArgs:
    def __init__(
        self,
        index_document: Optional[Input[str]] = None,
        sse_algorithm: Optional[Input[str]] = None,
        force_destroy: Optional[Input[bool]] = None,
    ):
        if index_document is not None:
            set_field(self, "index_document", index_document)
        if sse_algorithm is not None:
            set_field(self, "sse_algorithm", sse_algorithm)
        if force_destroy is not None:
            set_field(self, "force_destroy", force_destroy)

    @property
    def index_document(self) -> Optional[Input[str]]:
        return get_field(self, "index_document")

    @index_document.setter
    def index_document(self, value: Optional[Input[str]]) -> None:
        set_field(self, "index_document", value)

    @property
    def sse_algorithm(self) -> Optional[Input[str]]:
        return get_field(self, "sse_algorithm")

    @sse_algorithm.setter
    def sse_algorithm(self, value: Optional[Input[str]]) -> None:
        set_field(self, "sse_algorithm", value)

    @property
    def force_destroy(self) -> Optional[Input[bool]]:
        return get_field(self, "force_destroy")

    @force_destroy.setter
    def force_destroy(self, value: Optional[Input[bool]]) -> None:
        set_field(self, "force_destroy", value)


class Bucket(Construct):

    @overload
    def __init__(
        self, name: str, args: BucketArgs, opts: Optional[ConstructOptions] = None
    ): ...

    @overload
    def __init__(
        self,
        name,
        index_document: Optional[Input[str]] = None,
        sse_algorithm: Optional[Input[str]] = None,
        force_destroy: Optional[Input[bool]] = None,
        opts: Optional[ConstructOptions] = None,
    ): ...

    def __init__(self, name, *args, **kwargs):
        construct_args, opts = get_construct_args_opts(BucketArgs, *args, **kwargs)
        if construct_args is not None:
            self._internal_init(name, opts, **construct_args.__dict__)
        else:
            self._internal_init(name, *args, **kwargs)

    def _internal_init(
        self,
        name: str,
        opts: Optional[ConstructOptions] = None,
        index_document: Optional[Input[str]] = None,
        sse_algorithm: Optional[Input[str]] = None,
        force_destroy: Optional[Input[bool]] = None,
    ):
        super().__init__(
            name,
            construct_type="klotho.aws.Bucket",
            properties={
                "IndexDocument": index_document,
                "SseAlgorithm": sse_algorithm,
                "ForceDestroy": force_destroy,
            },
            opts=opts,
        )

    # Outputs
    @property
    def arn(self) -> Output[str]:
        return get_output(self, path="Arn", output_type=str)

    @property
    def bucket(self) -> Output[str]:
        return get_output(self, path="Bucket", output_type=str)

    # Bindings
    def use_read_only(self):
        """
        This method is used to create a binding for the bucket construct with read-only permissions.
        :return:
        """
        return Binding(self, inputs={"ReadOnly": True})

    def use_read_write(self):
        """
        This method is used to create a binding for the bucket construct with read-write permissions.
        :return: Binding
        """
        return Binding(self, inputs={"ReadOnly": False})



import re


class URN:
    def __init__(self, account_id="", project="", environment="", application="", type="", subtype="",
                 parent_resource_id="", resource_id="", output=""):
        self.account_id = account_id
        self.project = project
        self.environment = environment
        self.application = application
        self.type = type
        self.subtype = subtype
        self.parent_resource_id = parent_resource_id
        self.resource_id = resource_id
        self.output = output

    @classmethod
    def parse(cls, urn_string):
        matches = re.findall('[^:]+', urn_string)

        if len(matches) < 2:
            raise ValueError("Invalid URN format")

        if matches[0] == "urn":
            matches = matches[1:]

        urn = cls(account_id=matches[0], project=matches[1])

        if len(matches) > 2 and matches[2] != "":
            urn.environment = matches[2]
        if len(matches) > 3 and matches[3] != "":
            urn.application = matches[3]
        if len(matches) > 4 and matches[4] != "":
            type_parts = matches[4].split("/")
            if len(type_parts) != 2:
                raise ValueError("Invalid URN type format")
            urn.type = type_parts[0]
            urn.subtype = type_parts[1]

        if len(matches) > 5 and matches[5] != "":
            resource_parts = matches[5].split("/")
            if len(resource_parts) == 2:
                urn.parent_resource_id = resource_parts[0]
                urn.resource_id = resource_parts[1]
            else:
                urn.resource_id = matches[5]
        if len(matches) > 6 and matches[6] != "":
            urn.output = matches[6]

        if len(matches) > 7:
            raise ValueError("Invalid URN format")

        return urn

    def __str__(self):
        urn = f"urn:{self.account_id}:{self.project}:{self.environment}:{self.application}:"
        if self.type and self.subtype:
            urn += f"{self.type}/{self.subtype}:"
        if self.parent_resource_id and self.resource_id:
            urn += f"{self.parent_resource_id}/{self.resource_id}:"
        else:
            urn += f"{self.resource_id}:"
        urn += f"{self.output}:"

        # Remove trailing colons
        return urn.rstrip(":")

    def clone(self):
        return URN(
            account_id=self.account_id,
            project=self.project,
            environment=self.environment,
            application=self.application,
            type=self.type,
            subtype=self.subtype,
            parent_resource_id=self.parent_resource_id,
            resource_id=self.resource_id,
            output=self.output
        )

    def with_output(self, output: str):
        return URN(
            account_id=self.account_id,
            project=self.project,
            environment=self.environment,
            application=self.application,
            type=self.type,
            subtype=self.subtype,
            parent_resource_id=self.parent_resource_id,
            resource_id=self.resource_id,
            output=output
        )

    def __eq__(self, other):
        return str(self) == str(other)

    def __hash__(self):
        return hash(str(self))



from typing import Any, Optional, TypeVar, Type

from klotho import Output
from klotho.construct import Construct


def set_field(self, key: str, value: Any) -> None:
    """
    Set a value in the instance's dictionary.

    :param key: The key to set.
    :param value: The value to set.
    """
    self.__dict__[key] = value


def get_field(self, key: str) -> Optional[Any]:
    """
    Get a value from the instance's dictionary.

    :param key: The key to get.
    :return: The value, if it exists.
    """
    return self.__dict__.get(key)


T = TypeVar("T")


def get_output[T](self: Construct, path: str, output_type: Type[T] = any) -> Output[T]:
    """
    Get an output of the specified type from the resource at the specified path.

    :param self:
    :param output_type:  The type of the output.
    :param path: The path to the output.
    :return: The output.
    """

    output_urn = str(self.urn.with_output(path))

    return Output[output_type](
        id=output_urn,
        depends_on={output_urn},
    )



from typing import TYPE_CHECKING, Optional, Union, overload

from klotho.construct import (
    Binding,
    Construct,
    ConstructOptions,
    add_binding,
    get_construct_args_opts,
)
from klotho.output import Input, MappingInput, Output
from klotho.type_util import get_field, get_output, set_field

if TYPE_CHECKING:
    from klotho.aws.dynamodb import DynamoDB
    from klotho.aws.bucket import Bucket
    from klotho.aws.postgres import Postgres

BindingType = Union[
    Binding["DynamoDB"], "DynamoDB",
    Binding["Bucket"], "Bucket",
    Binding["Postgres"], "Postgres",
]

class FunctionArgs:
    def __init__(
        self,
        handler: Optional[Input[str]] = None,
        runtime: Optional[Input[str]] = None,
        timeout: Optional[Input[int]] = None,
        memory_size: Optional[Input[int]] = None,
        environment_variables: Optional[MappingInput[str]] = None,
        code: Optional[Input[str]] = None,
        s3_bucket: Optional[Input[str]] = None,
        s3_key: Optional[Input[str]] = None,
        s3_object_version: Optional[Input[str]] = None,
        image_uri: Optional[Input[str]] = None,
        dockerfile: Optional[Input[str]] = None,
        docker_context: Optional[Input[str]] = None,
        bindings: Optional[list[BindingType]] = None,
    ):
        if handler is not None:
            set_field(self, "handler", handler)
        if runtime is not None:
            set_field(self, "runtime", runtime)
        if timeout is not None:
            set_field(self, "timeout", timeout)
        if memory_size is not None:
            set_field(self, "memory_size", memory_size)
        if environment_variables is not None:
            set_field(self, "environment_variables", environment_variables)
        if code is not None:
            set_field(self, "code", code)
        if s3_bucket is not None:
            set_field(self, "s3_bucket", s3_bucket)
        if s3_key is not None:
            set_field(self, "s3_key", s3_key)
        if s3_object_version is not None:
            set_field(self, "s3_object_version", s3_object_version)
        if image_uri is not None:
            set_field(self, "image_uri", image_uri)
        if dockerfile is not None:
            set_field(self, "dockerfile", dockerfile)
        if docker_context is not None:
            set_field(self, "docker_context", docker_context)
        if bindings is not None:
            set_field(self, "bindings", bindings)

    @property
    def handler(self) -> Optional[Input[str]]:
        return get_field(self, "handler")

    @handler.setter
    def handler(self, value: Optional[Input[str]]) -> None:
        set_field(self, "handler", value)

    @property
    def runtime(self) -> Optional[Input[str]]:
        return get_field(self, "runtime")

    @runtime.setter
    def runtime(self, value: Optional[Input[str]]) -> None:
        set_field(self, "runtime", value)

    @property
    def timeout(self) -> Optional[Input[int]]:
        return get_field(self, "timeout")

    @timeout.setter
    def timeout(self, value: Optional[Input[int]]) -> None:
        set_field(self, "timeout", value)

    @property
    def memory_size(self) -> Optional[Input[int]]:
        return get_field(self, "memory_size")

    @memory_size.setter
    def memory_size(self, value: Optional[Input[int]]) -> None:
        set_field(self, "memory_size", value)

    @property
    def environment_variables(self) -> Optional[MappingInput[str]]:
        return get_field(self, "environment_variables")

    @environment_variables.setter
    def environment_variables(self, value: Optional[MappingInput[str]]) -> None:
        set_field(self, "environment_variables", value)

    @property
    def code(self) -> Optional[Input[str]]:
        return get_field(self, "code")

    @code.setter
    def code(self, value: Optional[Input[str]]) -> None:
        set_field(self, "code", value)

    @property
    def s3_bucket(self) -> Optional[Input[str]]:
        return get_field(self, "s3_bucket")

    @s3_bucket.setter
    def s3_bucket(self, value: Optional[Input[str]]) -> None:
        set_field(self, "s3_bucket", value)

    @property
    def s3_key(self) -> Optional[Input[str]]:
        return get_field(self, "s3_key")

    @s3_key.setter
    def s3_key(self, value: Optional[Input[str]]) -> None:
        set_field(self, "s3_key", value)

    @property
    def s3_object_version(self) -> Optional[Input[str]]:
        return get_field(self, "s3_object_version")

    @s3_object_version.setter
    def s3_object_version(self, value: Optional[Input[str]]) -> None:
        set_field(self, "s3_object_version", value)

    @property
    def image_uri(self) -> Optional[Input[str]]:
        return get_field(self, "image_uri")

    @image_uri.setter
    def image_uri(self, value: Optional[Input[str]]) -> None:
        set_field(self, "image_uri", value)

    @property
    def dockerfile(self) -> Optional[Input[str]]:
        return get_field(self, "dockerfile")

    @dockerfile.setter
    def dockerfile(self, value: Optional[Input[str]]) -> None:
        set_field(self, "dockerfile", value)

    @property
    def docker_context(self) -> Optional[Input[str]]:
        return get_field(self, "docker_context")

    @dockerfile.setter
    def docker_context(self, value: Optional[Input[str]]) -> None:
        set_field(self, "docker_context", value)


class Function(Construct):

    @overload
    def __init__(
        self, name: str, args: FunctionArgs, opts: Optional[ConstructOptions] = None
    ): ...

    @overload
    def __init__(
        self,
        name: str,
        handler: Optional[Input[str]] = None,
        runtime: Optional[Input[str]] = None,
        timeout: Optional[Input[int]] = None,
        memory_size: Optional[Input[int]] = None,
        environment_variables: Optional[MappingInput[str]] = None,
        code: Optional[Input[str]] = None,
        s3_bucket: Optional[Input[str]] = None,
        s3_key: Optional[Input[str]] = None,
        s3_object_version: Optional[Input[str]] = None,
        image_uri: Optional[Input[str]] = None,
        dockerfile: Optional[Input[str]] = None,
        docker_context: Optional[Input[str]] = None,
        bindings: Optional[list[BindingType]] = None,
        opts: Optional[ConstructOptions] = None,
    ): ...

    def __init__(self, name: str, *args, **kwargs):
        construct_args, opts = get_construct_args_opts(FunctionArgs, *args, **kwargs)
        if construct_args is not None:
            self._internal_init(name, opts=opts, **construct_args.__dict__)
        else:
            self._internal_init(name, *args, **kwargs)

    def _internal_init(
        self,
        name: str,
        handler: Optional[Input[str]] = None,
        runtime: Optional[Input[str]] = None,
        timeout: Optional[Input[int]] = None,
        memory_size: Optional[Input[int]] = None,
        environment_variables: Optional[MappingInput[str]] = None,
        code: Optional[Input[str]] = None,
        s3_bucket: Optional[Input[str]] = None,
        s3_key: Optional[Input[str]] = None,
        s3_object_version: Optional[Input[str]] = None,
        image_uri: Optional[Input[str]] = None,
        dockerfile: Optional[Input[str]] = None,
        docker_context: Optional[Input[str]] = None,
        bindings: Optional[list[BindingType]] = None,
        opts: Optional[ConstructOptions] = None,
    ):
        super().__init__(
            name,
            construct_type="klotho.aws.Function",
            properties={
                "Handler": handler,
                "Runtime": runtime,
                "Timeout": timeout,
                "MemorySize": memory_size,
                "EnvironmentVariables": (
                    Output.from_mapping(environment_variables)
                    if environment_variables
                    else None
                ),
                "Code": code,
                "S3Bucket": s3_bucket,
                "S3Key": s3_key,
                "S3ObjectVersion": s3_object_version,
                "ImageUri": image_uri,
                "Dockerfile": dockerfile,
                "DockerContext": docker_context,
            },
            bindings=bindings,
            opts=opts,
        )

    @property
    def function_arn(self) -> Output[str]:
        return get_output(self, "FunctionArn", str)

    @property
    def function_name(self) -> Output[str]:
        return get_output(self, "FunctionName", str)

    def bind(self, binding: BindingType) -> None:
        add_binding(self, binding)



from typing import Optional, overload

from klotho.construct import Construct, ConstructOptions, get_construct_args_opts
from klotho.output import Input
from klotho.type_util import set_field, get_field


class NetworkArgs:
    def __init__(self, name: Input[str]):
        set_field(self, "name", name)

    @property
    def name(self) -> Input[str]:
        return get_field(self, "name")

    @name.setter
    def name(self, value: Input[str]) -> None:
        set_field(self, "name", value)


class Network(Construct):

    @overload
    def __init__(self, args: NetworkArgs, opts: Optional[ConstructOptions] = None): ...

    @overload
    def __init__(self, name: Input[str], opts: Optional[ConstructOptions] = None): ...

    def __init__(self, *args, **kwargs):
        construct_args, opts = get_construct_args_opts(NetworkArgs, *args, **kwargs)
        if construct_args is not None:
            self._internal_init(opts, **construct_args.__dict__)
        else:
            self._internal_init(*args, **kwargs)

    def _internal_init(
        self,
        name: Input[str],
        opts: Optional[ConstructOptions] = None,
    ):
        super().__init__(
            name,
            construct_type="klotho.aws.Network",
            properties={},
            opts=opts,
        )



from typing import override, Type

from klotho.aws.network import Network
from klotho.provider import Provider, T
from klotho.runtime import instance as runtime


class AwsProvider(Provider):

    @override
    def get_default_construct(self, construct_type: Type[T]) -> T:
        if construct_type == Network:
            default = runtime.constructs.get("default-network")
            return default if default is not None else Network("default-network")
        raise ValueError(f"No default found for construct type: {construct_type}")



from typing import Optional, Type, TypeVar, Generic, Union, Any

from klotho.output import Output, Input
from klotho.runtime import instance as runtime
from klotho.urn import URN

BindingType = Union["Binding", "Construct"]


class ConstructOptions:
    """
    This class is a placeholder for when we need to add common options to constructs.
    """


class Construct:
    def __init__(
        self,
        name,
        construct_type,
        properties: dict[str, Any],
        bindings: Optional[list[BindingType]] = None,
        opts: Optional[ConstructOptions] = None,
    ):
        if runtime.application is None:
            raise Exception(
                "Cannot create a construct without an application. "
                "Initialize the application first by calling klotho.Application()."
            )

        self.urn = URN(
            account_id="my-account-id",  # TODO: Get this from the runtime or environment
            project=runtime.application.project,
            application=runtime.application.name,
            environment=runtime.application.environment,
            type="construct",
            subtype=construct_type,
            resource_id=name,
        )
        self.name = name
        self.construct_type = construct_type
        self.inputs = {}
        self.outputs = {}
        self.pulumi_stack = None
        self.version = 1
        self.status = "new"  # Default status
        self.bindings: list["Binding"] = []
        self.options = opts.__dict__ if opts else {}
        self.depends_on: set[URN] = set()
        runtime.add_construct(self)
        for k, v in properties.items():
            self.add_input(k, v)

        for binding in bindings or []:
            add_binding(self, binding)

    def to_dict(self):
        data = {
            "type": self.construct_type,
            "urn": str(self.urn),
            "version": self.version,
            "pulumi_stack": self.pulumi_stack,
            "status": self.status,
            "inputs": self.inputs,
            "outputs": self.outputs,
            "bindings": [b.to_dict() for b in self.bindings],
            "options": self.options,
            "dependsOn": [str(d) for d in self.depends_on],
        }
        return {k: v for k, v in data.items() if v}

    def add_input(self, name: str, value: Any):
        if value is not None:
            if isinstance(value, Construct):
                self.depends_on.add(value.urn)
                self.inputs[name] = {
                    "status": "pending",
                    "dependsOn": str(value.urn),
                }
            elif isinstance(value, Output):
                self.inputs[name] = {
                    "value": None,
                    "status": "pending",
                    "dependsOn": str(value.id),
                }

                for dep in {*value.depends_on, value.id}:
                    try:
                        # If the dep is a URN, add it to the list of dependencies (make sure to strip the output name).
                        urn = URN.parse(dep)
                        urn.output = ""
                        self.depends_on.add(urn)
                    except ValueError:
                        # If the dep is not a URN, it's not a construct dependency, so we can ignore it.
                        pass

            else:
                self.inputs[name] = {"value": value, "status": "resolved"}


def get_construct_args_opts(
    construct_args_type: Type, *args, **kwargs
) -> tuple[Any, ConstructOptions]:
    """
    Return the construct args and options given the *args and **kwargs of a construct's
    __init__ method.
    """

    construct_args, opts = None, None

    # If the first item is the construct args type, save it and remove it from the args list.
    if args and isinstance(args[0], construct_args_type):
        construct_args, args = args[0], args[1:]

    # Now look at the first item in the args list again.
    # If the first item is the construct options class, save it.
    if args and isinstance(args[0], ConstructOptions):
        opts = args[0]

    # If construct_args is None, see if "args" is in kwargs, and, if so, if it's typed as the
    # construct args type.
    if construct_args is None:
        a = kwargs.get("args")
        if isinstance(a, construct_args_type):
            construct_args = a

    # If opts is None, see if "opts" is in kwargs, and, if so, if it's an instance of ConstructOptions.
    if opts is None:
        o = kwargs.get("opts")
        opts = o if isinstance(o, ConstructOptions) else ConstructOptions()

    return construct_args, opts


TO = TypeVar("TO", bound=Construct)


class Binding(Generic[TO]):
    def __init__(self, to: TO, inputs: Optional[dict[str, Input[Any]]]):
        self._to: URN = to.urn
        self._inputs: dict[str, dict[str, Any]] = {}
        for k, v in inputs.items() if inputs else {}:
            self.add_input(k, v)

    def to_dict(self):
        return {"urn": str(self._to), "inputs": self._inputs}

    @property
    def to(self):
        return self._to

    @property
    def inputs(self):
        return {**self._inputs}

    def add_input(self, name: str, value: Any):
        if value is not None:
            if isinstance(value, Construct):
                self._inputs[name] = {
                    "status": "pending",
                    "dependsOn": str(value.urn),
                }
            elif isinstance(value, Output):
                self._inputs[name] = {
                    "value": None,
                    "status": "pending",
                    "dependsOn": str(value.id),
                }

            else:
                self._inputs[name] = {"value": value, "status": "resolved"}


def add_binding(source: Construct, binding: BindingType):
    # If the binding is a Construct, wrap it in a Binding.
    if isinstance(binding, Construct):
        binding = Binding(binding, {})

    if binding.to == source.urn:
        raise ValueError("Cannot bind a construct to itself.")

    # Replace any existing binding to the same construct.
    source.bindings = [b for b in source.bindings if b.to != binding.to]
    source.bindings.append(binding)

    source.depends_on.add(binding.to)

    # Add any input construct dependencies from the binding to the source.
    for v in binding.inputs.values():
        for dep in {*v.get("dependsOn", [])}:
            try:
                urn = URN.parse(dep)
                urn.output = ""
                source.depends_on.add(urn)
            except ValueError:
                # Ignore non-URN dependencies.
                pass



from typing import (
    TypeVar,
    Generic,
    Union,
    Mapping,
    Any,
    TYPE_CHECKING,
    Callable,
    Optional,
    overload,
)
from uuid import uuid4

from klotho.runtime import instance as runtime

if TYPE_CHECKING:
    pass

T = TypeVar("T")
K = TypeVar("K")
T_co = TypeVar("T_co", covariant=True)

Input = Union[T, "Output[T]"]
MappingInput = Union[Mapping[str, Input[T]], Input[Mapping[str, T]]]


class Output(Generic[T_co]):
    def __init__(
        self,
        depends_on: Optional[set[str]] = None,
        id: Optional[str] = None,
        value: Optional[T_co | Input[T_co]] = None,
        callback: Optional[Callable[[T_co], T]] = None,
    ):
        if id is not None:
            if id in runtime.output_references:
                raise ValueError(f"Output with id {id} already exists")
            self.id = id
        else:
            self.id = str(uuid4())
        self.depends_on = set(depends_on) if depends_on else set()

        if isinstance(value, Output) and value:
            self._value = value.value
            self._is_resolved = value.is_resolved
        self._value = value
        self._is_resolved = value is not None
        self.callback = callback

        runtime.output_references[self.id] = self

    @property
    def is_resolved(self) -> bool:
        return self._is_resolved

    @property
    def value(self) -> T_co:
        if not self._is_resolved:
            raise ValueError("Output value is not resolved")
        return self._value

    def __str__(self):
        return f"Output({self.id}={self._value})"

    def apply(self, callback: Callable[[T_co], T]) -> "Output[T]":
        def run(value: T_co) -> T_co:
            if self._is_resolved:
                return self._value
            self._value = callback(value)
            self._is_resolved = True
            return self._value

        return Output(self.depends_on, self._value, run)

    def resolve(self, resolved_deps):
        if self._is_resolved:
            return
        self._value = (
            self.callback(*resolved_deps) if self.callback else resolved_deps[0]
        )
        self._is_resolved = True
        runtime.outputs[self.id] = self._value

    @staticmethod
    @overload
    def all(
        outputs: list["Output[T_co]"], callback: Callable[..., T_co] | None = None
    ) -> "Output[T_co]": ...

    @staticmethod
    @overload
    def all(
        outputs: Mapping[K, "Output[T_co]"],
        callback: Callable[[Mapping[K, T_co]], T_co] | None = None,
    ) -> "Output[T_co]": ...

    @staticmethod
    def all(
        outputs: list["Output[T_co]"] | Mapping[Any, "Output[T_co]"],
        callback: (
            Callable[..., T_co] | Callable[[Mapping[Any, T_co]], T_co] | None
        ) = None,
    ) -> "Output[T_co]":
        """
        Creates a new Output that represents the output of applying a callback function to the given outputs.
        Accepts either a list of Output objects or a Mapping of keys to Output objects.
        :param outputs: List of Output objects or Mapping of keys to Output objects
        :param callback: Callback function to apply to the outputs
        :return: The new Output
        """
        if isinstance(outputs, Mapping):

            output_list = []
            output_keys = []
            for key, output in outputs.items():
                output_list.append(output)
                output_keys.append(key)

            def run(*values: T_co) -> T_co:
                resolved_outputs = {
                    key: value for key, value in zip(output_keys, values)
                }
                return callback(resolved_outputs)  # type: ignore

        else:

            def run(*values: T_co) -> T_co:
                return callback(*values)  # type: ignore

            output_list = outputs
            output_keys = None

        return Output(
            depends_on={
                *[output.id for output in output_list],
                *[dep for output in output_list for dep in output.depends_on],
            },
            callback=run,
        )

    @classmethod
    def concat(cls, *args: "Input[str]") -> "Output[str]":
        """
        Concatenates the string representations of all the given inputs.
        :param args: The inputs to concatenate.
        :return: A new Output representing the concatenated string.
        """
        inputs = [
            arg if isinstance(arg, Output) else Output(set(), None, arg) for arg in args
        ]

        def run(*values: str) -> str:
            return "".join(values)

        return cls.all(inputs, run)

    @staticmethod
    def from_mapping(input: Input[Mapping]) -> "Output[Mapping]":
        if isinstance(input, Output):
            return input
        if isinstance(input, Mapping):
            deps = set()
            unresolved_mappings = {}
            resolved_mappings = {}
            for key, value in input.items():
                if isinstance(value, Output):
                    deps.update(value.depends_on)
                    unresolved_mappings[key] = value
                else:
                    resolved_mappings[key] = value

            def callback(resolved_outputs: Mapping) -> Mapping:
                result = {**resolved_mappings, **resolved_outputs}
                return result

            return Output.all(
                unresolved_mappings,
                callback=callback,
            )
        else:
            raise ValueError("Input must be an Output or a Mapping")



from typing import Optional

from klotho.construct import Binding, Construct, ConstructOptions, add_binding


class Api(Construct):
    def __init__(self, name: str, opts: Optional[ConstructOptions] = None):
        super().__init__(
            name, construct_type="klotho.aws.Api", properties={}, opts=opts
        )

    def route_to(self, path: str, dest: Construct):
        add_binding(self, Binding(dest, {"Path": path}))



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: service.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\rservice.proto\x12\x06klotho\"\x1d\n\tIRRequest\x12\x10\n\x08\x66ilename\x18\x01 \x01(\t\"0\n\x07IRReply\x12\x0f\n\x07message\x18\x01 \x01(\t\x12\x14\n\x0cyaml_payload\x18\x02 \x01(\t\"\x14\n\x12HealthCheckRequest\"\"\n\x10HealthCheckReply\x12\x0e\n\x06status\x18\x01 \x01(\t\"0\n\x18RegisterConstructRequest\x12\x14\n\x0cyaml_payload\x18\x01 \x01(\t\"?\n\x16RegisterConstructReply\x12\x0f\n\x07message\x18\x01 \x01(\t\x12\x14\n\x0cyaml_payload\x18\x02 \x01(\t2\xdf\x01\n\rKlothoService\x12.\n\x06SendIR\x12\x11.klotho.IRRequest\x1a\x0f.klotho.IRReply\"\x00\x12\x45\n\x0bHealthCheck\x12\x1a.klotho.HealthCheckRequest\x1a\x18.klotho.HealthCheckReply\"\x00\x12W\n\x11RegisterConstruct\x12 .klotho.RegisterConstructRequest\x1a\x1e.klotho.RegisterConstructReply\"\x00\x42:Z8github.com/klothoplatform/klotho/pkg/k2/language_host/gob\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'service_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'Z8github.com/klothoplatform/klotho/pkg/k2/language_host/go'
  _globals['_IRREQUEST']._serialized_start=25
  _globals['_IRREQUEST']._serialized_end=54
  _globals['_IRREPLY']._serialized_start=56
  _globals['_IRREPLY']._serialized_end=104
  _globals['_HEALTHCHECKREQUEST']._serialized_start=106
  _globals['_HEALTHCHECKREQUEST']._serialized_end=126
  _globals['_HEALTHCHECKREPLY']._serialized_start=128
  _globals['_HEALTHCHECKREPLY']._serialized_end=162
  _globals['_REGISTERCONSTRUCTREQUEST']._serialized_start=164
  _globals['_REGISTERCONSTRUCTREQUEST']._serialized_end=212
  _globals['_REGISTERCONSTRUCTREPLY']._serialized_start=214
  _globals['_REGISTERCONSTRUCTREPLY']._serialized_end=277
  _globals['_KLOTHOSERVICE']._serialized_start=280
  _globals['_KLOTHOSERVICE']._serialized_end=503
# @@protoc_insertion_point(module_scope)



import threading
from typing import Optional, TYPE_CHECKING, Any

import grpc
import yaml

import klotho
from klotho import service_pb2_grpc as service_pb2_grpc
from klotho.provider import Provider
from klotho.urn import URN

if TYPE_CHECKING:
    from klotho.output import Output


class Runtime:
    _instance = None
    _lock = threading.Lock()

    def __new__(cls):
        if cls._instance is None:
            with cls._lock:
                if cls._instance is None:
                    cls._instance = super(Runtime, cls).__new__(cls)
                    cls._instance._initialized = False
        return cls._instance

    def __init__(self):
        if not self._initialized:
            self.constructs = {}
            self.outputs: dict[str, Any] = {}
            self.output_references: dict[str, "Output"] = {}
            self.application: Optional["klotho.Application"] = None
            channel = grpc.insecure_channel("localhost:50051")
            self.stub = service_pb2_grpc.KlothoServiceStub(channel)
            self._initialized = True
            self.providers: dict[str, Provider] = {}

    def add_construct(self, resource):
        self.constructs[resource.name] = resource

    def set_application(self, application):
        self.application = application

    def generate_yaml(self):
        constructs = {
            name: construct.to_dict() for name, construct in self.constructs.items()
        }
        output = {
            "schemaVersion": 1,  # Adjust this value as needed
            "version": 1,  # Adjust this value as needed
            "app_urn": f"urn:accountid:{self.application.project}:{self.application.environment}:{self.application.name}",
            "project_urn": f"urn:accountid:{self.application.project}",
            "environment": self.application.environment,
            "default_region": self.application.default_region,
            "constructs": constructs,
        }
        return yaml.dump(output, sort_keys=False)

    def resolve_output_references(
        self, constructs: dict[str, dict[str, Any]]
    ) -> dict[str, Any]:
        """
        constructs is expected to be a dictionary of resource urn to resource outputs
        example:
        {
            "urn:accountid:project:env:construct:...": {
                "output1": "value1",
                "output2": "value2"
            }
        }

        :return: a dictionary of resolved output references where the key is the output id and the value is the resolved value
        """
        for urn, outputs in constructs.items():
            for output_name, output_value in outputs.items():
                output_urn = URN.parse(urn)
                output_urn.output = output_name
                self.outputs[str(output_urn)] = output_value

        remaining_unresolved_outputs = {
            output_id: output
            for output_id, output in self.output_references.items()
            if not output.is_resolved
        }
        resolved_output_references = {}
        resolved_count = -1
        while resolved_count != 0:
            resolved_count = 0
            unresolved_outputs = remaining_unresolved_outputs
            remaining_unresolved_outputs = {}
            for output_id, output in unresolved_outputs.items():
                resolved_deps = []
                for dep in output.depends_on:
                    if dep in self.outputs:
                        resolved_deps.append(self.outputs[dep])
                if len(resolved_deps) != len(output.depends_on):
                    remaining_unresolved_outputs[output_id] = output
                    continue
                output.resolve(resolved_deps)
                resolved_count += 1
                resolved_output_references[output_id] = output.value
        return resolved_output_references

    def set_provider(self, name: str, provider: Provider):
        self.providers[name] = provider


instance: Runtime = Runtime()



from typing import TypeVar, Type

from klotho.construct import Construct
from klotho.runtime import instance as runtime

Tc = TypeVar('Tc', bound=Construct)


def get_default_construct(provider: str, construct_type: Type[Tc]) -> Tc:
    provider = runtime.providers.get(provider)
    if provider is None:
        raise ValueError(f"Provider {provider} not found")
    return provider.get_default_construct(construct_type)



from typing import Optional, overload

from klotho.aws.network import Network
from klotho.construct import ConstructOptions, get_construct_args_opts, Construct
from klotho.output import Input
from klotho.runtime_util import get_default_construct
from klotho.type_util import set_field, get_field


class PostgresArgs:
    """Arguments for configuring a Postgres database."""

    def __init__(
                 self,
                 database_name: Input[str],
                 instance_class: Optional[Input[str]] = None,
                 allocated_storage: Optional[Input[int]] = None,
                 engine_version: Optional[Input[str]] = None,
                 username: Optional[Input[str]] = None,
                 password: Optional[Input[str]] = None,
                 port: Optional[Input[int]] = None,
                 network: Optional[Network] = None
                ):
        if instance_class is not None:
            set_field(self, "instance_class", instance_class)
        if allocated_storage is not None:
            set_field(self, "allocated_storage", allocated_storage)
        if engine_version is not None:
            set_field(self, "engine_version", engine_version)
        if username is not None:
            set_field(self, "username", username)
        if password is not None:
            set_field(self, "password", password)
        if database_name is not None:
            set_field(self, "database_name", database_name)
        if port is not None:
            set_field(self, "port", port)
        if network is not None:
            set_field(self, "network", network)

    def _get_property(self, name: str):
        return get_field(self, name)

    def _set_property(self, name: str, value):
        set_field(self, name, value)

    @property
    def instance_class(self) -> Optional[Input[str]]:
        return self._get_property("instance_class")

    @instance_class.setter
    def instance_class(self, value: Optional[Input[str]]) -> None:
        self._set_property("instance_class", value)

    @property
    def allocated_storage(self) -> Optional[Input[int]]:
        return self._get_property("allocated_storage")

    @allocated_storage.setter
    def allocated_storage(self, value: Optional[Input[int]]) -> None:
        self._set_property("allocated_storage", value)

    @property
    def engine_version(self) -> Optional[Input[str]]:
        return self._get_property("engine_version")

    @engine_version.setter
    def engine_version(self, value: Optional[Input[str]]) -> None:
        self._set_property("engine_version", value)

    @property
    def username(self) -> Optional[Input[str]]:
        return self._get_property("username")

    @username.setter
    def username(self, value: Optional[Input[str]]) -> None:
        self._set_property("username", value)

    @property
    def password(self) -> Optional[Input[str]]:
        return self._get_property("password")

    @password.setter
    def password(self, value: Optional[Input[str]]) -> None:
        self._set_property("password", value)

    @property
    def database_name(self) -> Optional[Input[str]]:
        return self._get_property("database_name")

    @database_name.setter
    def database_name(self, value: Optional[Input[str]]) -> None:
        self._set_property("database_name", value)

    @property
    def port(self) -> Optional[Input[int]]:
        return self._get_property("port")

    @port.setter
    def port(self, value: Optional[Input[int]]) -> None:
        self._set_property("port", value)

    @property
    def network(self) -> Network:
        return self._get_property("network")

    @network.setter
    def network(self, value: Optional[Network]) -> None:
        self._set_property("network", value)


class Postgres(Construct):
    """Represents a Postgres database construct in AWS."""

    @overload
    def __init__(
        self, name: str, args: PostgresArgs, opts: Optional[ConstructOptions] = None
    ): ...

    @overload
    def __init__(
        self,
        name: str,
        instance_class: Optional[Input[str]] = None,
        allocated_storage: Optional[Input[int]] = None,
        engine_version: Optional[Input[str]] = None,
        username: Optional[Input[str]] = None,
        password: Optional[Input[str]] = None,
        database_name: Optional[Input[str]] = None,
        port: Optional[Input[int]] = None,
        network: Optional[Network] = None,
        opts: Optional[ConstructOptions] = None,
    ): ...

    def __init__(self, name: str, *args, **kwargs):
        construct_args, opts = get_construct_args_opts(PostgresArgs, *args, **kwargs)
        if construct_args is not None:
            self._internal_init(name, opts, **construct_args.__dict__)
        else:
            self._internal_init(name, *args, **kwargs)

    def _internal_init(
        self,
        name: str,
        instance_class: Optional[Input[str]] = None,
        allocated_storage: Optional[Input[int]] = None,
        opts: Optional[ConstructOptions] = None,
        engine_version: Optional[Input[str]] = None,
        username: Optional[Input[str]] = None,
        password: Optional[Input[str]] = None,
        database_name: Optional[Input[str]] = None,
        port: Optional[Input[int]] = None,
        network: Optional[Network] = None,
    ):
        """Internal initializer for Postgres."""
        if network is None:
            network = get_default_construct("aws", Network)

        if instance_class is None:
            instance_class = "db.t3.micro"
        if allocated_storage is None:
            allocated_storage = 20

        super().__init__(
            name,
            construct_type="klotho.aws.Postgres",
            properties={
                "InstanceClass": instance_class,
                "AllocatedStorage": allocated_storage,
                "EngineVersion": engine_version,
                "Username": username,
                "Password": password,
                "DatabaseName": database_name,
                "Port": port,
                "Network": network,
            },
            opts=opts,
        )

    @property
    def endpoint(self) -> str:
        """The endpoint of the Postgres database."""
        return get_field(self, "Endpoint")

    @property
    def port(self) -> int:
        """The port of the Postgres database."""
        return get_field(self, "Port")



import os

from klotho.runtime import instance as runtime


class Application:
    def __init__(self, name, project=None, environment=None, default_region=None):
        self.name = name
        self.project = project or os.getenv('PROJECT_NAME', 'default')
        self.environment = environment or os.getenv('KLOTHO_ENVIRONMENT', 'default')
        self.default_region = default_region or os.getenv('AWS_REGION', 'us-east-1')
        runtime.set_application(self)



from klotho.aws.api import Api
from klotho.aws.bucket import Bucket
from klotho.aws.container import Container
from klotho.aws.postgres import Postgres
from klotho.aws.dynamodb import DynamoDB
from klotho.aws.fastapi import FastAPI
from klotho.aws.function import Function

from klotho.aws.provider import AwsProvider
from klotho.runtime import instance as runtime

runtime.set_provider("aws", AwsProvider())
